"""Query Firestore collections."""
import subprocess  # noqa: S404
from typing import List

from .utils import drop_empty


def get_collections() -> List[str]:
    """Get collections from a database."""
    output = subprocess.check_output(["fuego", "c"])  # noqa: S603, S607
    return drop_empty(output.decode("utf-8").split("\n"))



"""Query and return ROAR runs."""
from datetime import date
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pandas as pd
from dateutil.parser import isoparse
from tqdm.auto import tqdm

from .utils import _FuegoKey
from .utils import _FuegoResponse
from .utils import page_results
from .utils import trim_doc_path


def merge_data_with_metadata(
    fuego_response: List[_FuegoResponse], metadata_params: Dict[str, _FuegoKey]
) -> List[Dict[str, Any]]:
    """Merge trial data with metadata.

    We often want to merge the run/trial data with some of the metadata returned
    by Firestore. In Python 3.9 we could use the | operator but we want to be
    backward compatible so we iterate over the data and metadata and merge them
    together.

    Parameters
    ----------
    fuego_response : List[Dict[str, Any]]
        The trial data.

    metadata_params : Dict[str, str]
        The metadata fields that will be merged into the data. The keys are the
        desired keys in the merged data and the values are the metadata keys to
        use.

    Returns
    -------
    List[Dict[str, Any]]
        The merged data.
    """
    item_data = [item["Data"] for item in fuego_response]
    for item, raw_item in zip(item_data, fuego_response):
        item.update(
            {
                data_key: raw_item[metadata_key]
                for data_key, metadata_key in metadata_params.items()
            }
        )

    return item_data


def get_trials_from_run(run_path: str) -> List[Dict[str, Any]]:
    """Get all trials from a run.

    Parameters
    ----------
    run_path : str
        The Firestore path to the run.

    Returns
    -------
    List[Dict[str, str]]
        The trials from the run.
    """
    trial_path = f"{trim_doc_path(run_path)}/trials"
    fuego_query = ["fuego", "query", trial_path]
    raw_trials = page_results(fuego_query)

    return merge_data_with_metadata(
        fuego_response=raw_trials,
        metadata_params={"CreateTime": "CreateTime", "trialId": "ID"},
    )


def filter_run_dates(
    runs: List[_FuegoResponse],
    started_before: Optional[Union[date, datetime]] = None,
    started_after: Optional[Union[date, datetime]] = None,
) -> List[_FuegoResponse]:
    """Filter runs by date.

    Parameters
    ----------
    runs : List[Dict[str, Any]]
        The runs to filter.

    started_before : date, optional, default=None
        Return only runs started before this date.

    started_after : date, optional, default=None
        Return only runs started after this date.

    Returns
    -------
    List[Dict[str, Any]]
        The filtered runs.

    Examples
    --------
    >>> runs = [
    ...     {
    ...         "CreateTime": "2020-01-01T00:00:00.000Z",
    ...         "Data": {
    ...             "name": "run-1",
    ...             "timeStarted": "2020-01-01T00:00:00.000Z",
    ...             "classId": "class-1",
    ...             "completed": "true",
    ...         },
    ...         "ID": "run-1",
    ...         "Path": "prod/roar-prod/runs/run-1",
    ...         "ReadTime": "2020-01-01T00:00:00.000Z",
    ...         "UpdateTime": "2020-01-01T00:00:00.000Z",
    ...     },
    ...     {
    ...         "CreateTime": "2020-02-01T00:00:00.000Z",
    ...         "Data": {
    ...             "name": "run-2",
    ...             "timeStarted": "2020-02-01T00:00:00.000Z",
    ...             "classId": "class-2",
    ...             "completed": "true",
    ...         },
    ...         "ID": "run-2",
    ...         "Path": "prod/roar-prod/runs/run-1",
    ...         "ReadTime": "2020-02-01T00:00:00.000Z",
    ...         "UpdateTime": "2020-02-01T00:00:00.000Z",
    ...     },
    ... ]
    >>> filtered = filter_run_dates(runs, started_before=date(2020, 1, 15))
    >>> print(filtered == [runs[0]])
    True
    """
    if isinstance(started_before, date):
        started_before = datetime(
            started_before.year, started_before.month, started_before.day
        ).astimezone()

    if isinstance(started_after, date):
        started_after = datetime(
            started_after.year, started_after.month, started_after.day
        ).astimezone()

    filtered = [run for run in runs]

    if started_before is not None:
        filtered = [
            run
            for run in filtered
            if isoparse(run["Data"]["timeStarted"]) < started_before
        ]

    if started_after is not None:
        filtered = [
            run
            for run in filtered
            if isoparse(run["Data"]["timeStarted"]) > started_after
        ]

    return filtered


def get_runs(
    root_doc: str = "prod/roar-prod",
    return_trials: bool = False,
    query_kwargs: Optional[Dict[str, str]] = None,
    started_before: Optional[date] = None,
    started_after: Optional[date] = None,
) -> pd.DataFrame:
    """Get all runs that satisfy a specific query.

    Parameters
    ----------
    root_doc : str, optional, default="prod/roar-prod"
        The Firestore root document. The returned runs will all be under this document.

    return_trials : bool, optional, default=False
        If True, return the trials for each run as well.

    query_kwargs : dict, optional, default=None
        The query to run. If None, all runs will be returned.

    started_before : date, optional, default=None
        Return only runs started before this date.

    started_after : date, optional, default=None
        Return only runs started after this date.

    Returns
    -------
    List[dict]
        The runs that satisfy the query.
    """
    # Build the fuego query dynamically
    fuego_args = ["fuego", "query"]
    for select in [
        "classId",
        "completed",
        "districtId",
        "schoolId",
        "studyId",
        "taskId",
        "timeFinished",
        "timeStarted",
        "variantId",
    ]:
        fuego_args.extend(["--select", select])

    query_kwargs = query_kwargs if query_kwargs is not None else {}

    # Treat the roar UID separately
    roar_uid = query_kwargs.pop("roarUid", None)
    roar_uid_prefix = query_kwargs.pop("roarUidPrefix", None)

    if roar_uid is None:
        query = ["-g", "runs"]
    else:
        query = ["/".join([root_doc.rstrip("/"), "users", roar_uid, "runs"])]

    for key, value in query_kwargs.items():
        query.append(f'{key} == "{value}"')

    fuego_args.extend(query)
    runs = page_results(fuego_args)

    if not runs:
        raise ValueError("Your query returned no results.")

    # Get rid of results that are not in the root_doc
    runs = [run for run in runs if root_doc in run["Path"]]

    if roar_uid_prefix is not None:
        # Get rid of results that do not have the UID prefix
        roar_uid_prefix = "/".join(
            [root_doc.rstrip("/"), "users", roar_uid_prefix.strip("/")]
        )
        runs = [run for run in runs if roar_uid_prefix in run["Path"]]

    # Get rid of runs that are outside of the date range
    runs = filter_run_dates(
        runs=runs, started_before=started_before, started_after=started_after
    )

    df_runs = pd.DataFrame(
        merge_data_with_metadata(
            fuego_response=runs,
            metadata_params={"CreateTime": "CreateTime", "runId": "ID"},
        )
    )

    df_runs.set_index("runId", inplace=True)

    if not return_trials:
        return df_runs

    run_paths = {run["ID"]: run["Path"] for run in runs}

    run_trials = {
        run_id: get_trials_from_run(run_path)
        for run_id, run_path in tqdm(run_paths.items(), desc="Getting trials")
    }

    run_trials = {
        run_id: pd.DataFrame(trials) for run_id, trials in run_trials.items() if trials
    }

    for run_id, df in run_trials.items():
        df["runId"] = run_id  # type: ignore [call-overload]

    df_trials = pd.concat(run_trials.values())

    df_trials.set_index("trialId", inplace=True)
    df_trials = df_trials.merge(df_runs, left_on="runId", right_index=True, how="left")

    return df_trials



"""Command-line interface."""
from datetime import date

import click

from .runs import get_runs
from .utils import camel_case


@click.version_option()
@click.group(
    epilog="""
\b
Useful definitions:
- trial: a single stimulus/response pair
- run: globally unique collection of successive trials.
       This constitute “running” through the task one time.
- corpus: A named and immutable collection of stimuli
- block: A portion of a run whose stimuli are drawn from only one corpus.
- task: the activity or assessment that was performed in a run (e.g. SWR, PA, SRE, etc.)
- task-variant: a specification of the task, e.g. adaptive vs. random; 1 vs 3 blocks
- school, district, class: all assume the standard meaning
- study: a collection of runs associated with a research project
"""
)
def main() -> None:
    """Roarquery.

    Roarquery is a command-line interface for querying ROAR data in Google Cloud
    Firestore. It has several subcommands, which are listed below.
    """
    pass


@main.command(
    epilog="""
Examples:

  Return trials for the "swr" task in the "validation" study.

    $ roarquery runs --task-id=swr --study-id=validation --return-trials trials.csv

  Return runs for the "sre" task in the "sd" district that started after 2021-05-10.

    $ roarquery runs --task-id=sre --district-id=sd --started-after=2021-05-10 runs.csv
"""
)
@click.option(
    "--roar-uid", type=str, help="Return only runs for the user with this ROAR UID."
)
@click.option(
    "--roar-uid-prefix", type=str, help="Return only runs for users with this prefix."
)
@click.option("--task-id", type=str, help="Return only runs for this task.")
@click.option("--study-id", type=str, help="Return only runs for this study.")
@click.option("--variant-id", type=str, help="Return only runs for this variant.")
@click.option("--district-id", type=str, help="Return only runs for this district.")
@click.option("--school-id", type=str, help="Return only runs for this school.")
@click.option("--class-id", type=str, help="Return only runs with this class.")
@click.option(
    "--require-completed",
    is_flag=True,
    show_default=True,
    default=False,
    help="Require all runs to be completed.",
)
@click.option(
    "--started-before",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Return only runs started before this date. Format: YYYY-MM-DD.",
)
@click.option(
    "--started-after",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="Return only runs started after this date. Format: YYYY-MM-DD.",
)
@click.option(
    "--return-trials",
    is_flag=True,
    default=False,
    help="Return the trials for each run as well.",
)
@click.option(
    "--root-doc",
    type=str,
    default="prod/roar-prod",
    help="The Firestore root document. Returned runs will all be under this document.",
)
@click.argument(
    "output_filename",
    type=click.Path(dir_okay=False, writable=True),
)
def runs(
    roar_uid: str,
    roar_uid_prefix: str,
    task_id: str,
    study_id: str,
    variant_id: str,
    district_id: str,
    school_id: str,
    class_id: str,
    require_completed: bool,
    started_before: date,
    started_after: date,
    return_trials: bool,
    root_doc: str,
    output_filename: click.Path,
) -> None:
    r"""Return ROAR runs matching certain query parameters.

    The options described below can be combined to return runs that match
    all of the specified query parameters.

    \b
    Arguments:
      OUTPUT FILENAME            Path to the output file to which to save runs/trials.
    """
    query_kwargs = {
        "roar_uid": roar_uid,
        "roar_uid_prefix": roar_uid_prefix,
        "task_id": task_id,
        "study_id": study_id,
        "variant_id": variant_id,
        "district_id": district_id,
        "school_id": school_id,
        "class_id": class_id,
    }

    # Convert to camelCase and remove None values.
    query_kwargs = {
        camel_case(key): value
        for key, value in query_kwargs.items()
        if value is not None
    }

    if require_completed:
        query_kwargs["completed"] = "true"

    df_trials = get_runs(
        root_doc=root_doc,
        return_trials=return_trials,
        query_kwargs=query_kwargs,
        started_before=started_before,
        started_after=started_after,
    )

    df_trials.to_csv(output_filename, index=True)


if __name__ == "__main__":
    main(prog_name="roarquery")  # pragma: no cover



"""Utilities functions."""
import json
import subprocess  # noqa: S404
from re import sub
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import TypedDict


_FuegoKey = Literal["CreateTime", "Data", "ID", "Path", "ReadTime", "UpdateTime"]


class _FuegoResponse(TypedDict):
    CreateTime: str
    Data: Dict[str, Any]
    ID: str
    Path: str
    ReadTime: str
    UpdateTime: str


def camel_case(string: str) -> str:
    """Convert a string to camel case.

    Parameters
    ----------
    string : str
        The string to convert.

    Returns
    -------
    str
        The camel case string.

    Examples
    --------
    >>> camel_case("an_example_string")
    anExampleString

    >>> camel_case("an-example-string-with-dashes")
    anExampleStringWithDashes
    """
    string = sub(r"(_|-)+", " ", string).title().replace(" ", "")
    return "".join([string[0].lower(), string[1:]])


def bytes2json(bytes: bytes) -> List[_FuegoResponse]:
    r"""Convert bytes to json.

    Parameters
    ----------
    bytes : bytes
        The bytes to convert.

    Returns
    -------
    List[_FuegoResponse]
        The converted json.

    Examples
    --------
    An empty string will return an empty list:
    >>> bytes2json(b"")
    []

    >>> json_out = bytes2json(
    ...    b'[{"ID": "1", "Data": {"a": "b"}, "Path": "prod/roar-prod", '
    ...    b'"CreateTime": "2020-04-01T00:00:00Z", '
    ...    b'"ReadTime": "2020-04-01T00:00:00Z", '
    ...    b'"UpdateTime": "2020-04-01T00:00:00Z"}]'
    ... )
    >>> print(json_out == [{
    ...     'ID': '1',
    ...     'Data': {'a': 'b'},
    ...     'Path': 'prod/roar-prod',
    ...     'CreateTime': '2020-04-01T00:00:00Z',
    ...     'ReadTime': '2020-04-01T00:00:00Z',
    ...     'UpdateTime': '2020-04-01T00:00:00Z'
    ... }])
    True
    """
    if not bytes:
        return []
    return cast(List[_FuegoResponse], json.loads(bytes.decode("utf-8")))


def page_results(query: List[str], limit: Optional[int] = None) -> List[_FuegoResponse]:
    """Page through results from a query.

    Parameters
    ----------
    query : List[str]
        The query to run. This is a list of strings that will be passed to
        subprocess.check_output.

    limit : int, optional, default=100
        The number of results to return per page.

    Returns
    -------
    List[_FuegoResponse]
        The results of the query.
    """
    limit = limit if limit is not None else 100
    query_idx = query.index("query")
    query.insert(query_idx + 1, "--limit")
    query.insert(query_idx + 2, str(limit))

    output = []
    this_page = subprocess.check_output(query)  # noqa: S603

    while this_page:
        output.extend(bytes2json(this_page))

        if len(bytes2json(this_page)) == limit:
            if "--startafter" in query:
                start_after_idx = query.index("--startafter")
                query[start_after_idx + 1] = trim_doc_path(output[-1]["Path"])
            else:
                query.insert(query_idx + 1, "--startafter")
                query.insert(query_idx + 2, trim_doc_path(output[-1]["Path"]))

            this_page = subprocess.check_output(query)  # noqa: S603
        else:
            this_page = b""

    return output


def drop_empty(iterable: List[Any]) -> List[Any]:
    """Drop empty strings from a list.

    Parameters
    ----------
    iterable : list
        The list to drop empty strings from.

    Returns
    -------
    list
        The list with empty strings dropped.

    Examples
    --------
    >>> drop_empty(["", "a", "", "b"])
    ['a', 'b']
    """
    return [x for x in iterable if x]


def trim_doc_path(path: str) -> str:
    """Remove leading project information from firestore document path.

    Parameters
    ----------
    path : str
        The path to standardize.

    Returns
    -------
    str
        The standardized path.

    Examples
    --------
    >>> trim_doc_path("projects/proj-id/databases/(default)/documents/prod/roar-prod")
    prod/roar-prod
    """
    return path.split("databases/(default)/documents/")[-1]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['roarquery']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'pandas>=1.4.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['roarquery = roarquery.__main__:main']}

setup_kwargs = {
    'name': 'roarquery',
    'version': '0.2.1',
    'description': 'Roarquery',
    'long_description': 'Roarquery\n=========\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/roarquery.svg\n   :target: https://pypi.org/project/roarquery/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/roarquery.svg\n   :target: https://pypi.org/project/roarquery/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/roarquery\n   :target: https://pypi.org/project/roarquery\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/roarquery\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/roarquery/latest.svg?label=Read%20the%20Docs\n   :target: https://roarquery.readthedocs.io/\n   :alt: Read the documentation at https://roarquery.readthedocs.io/\n.. |Tests| image:: https://github.com/richford/roarquery/workflows/Tests/badge.svg\n   :target: https://github.com/richford/roarquery/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/richford/roarquery/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/richford/roarquery\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n\n* Query ROAR runs\n* Download ROAR runs and trials\n* List ROAR Firestore collections\n\n\nRequirements\n------------\n\n* Python 3.8+\n* `fuego`_\n\n\nInstallation\n------------\n\nYou can install *Roarquery* via pip_ from PyPI_:\n\n.. code:: console\n\n   pip install roarquery\n\n*Roarquery* also requires you to install *fuego*, a command line firestore client.\nPlease see the `fuego documentation`_ for complete installation instructions.\n\nOn a Mac, follow these steps:\n\n1. Ensure you have a working go installation. If\n\n.. code:: console\n\n   go version\n\nreturns something, then you are good to go. If not, install go with homebrew:\n\n.. code:: console\n\n   /bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh)"\n   brew install go\n\n2. Then install *fuego*\n\n.. code:: console\n\n   git clone https://github.com/sgarciac/fuego.git\n   cd fuego\n   go build .\n   go install .\n\n3. Finally, modify your PATH variable to include the go installation directory, which can be done with the following incantation:\n\n.. code:: console\n\n   echo $HOME/go/bin | sudo tee -a /private/etc/paths.d/go\n\n4. You may need to open a new terminal window or tab for these changes to take effect.\n\nUsage\n-----\n\nAuthentication\n~~~~~~~~~~~~~~\n\nBefore you can use *Roarquery*, you need to provide authentication details:\n\n1. Retrieve or generate a Service Account key file.\n\n   a. go to your `Firebase project console`_,\n   b. go to "Project settings" (in the little gear menu next to "Project Overview"),\n   c. click on the "Service accounts" tab,\n   d. click on the "Generate new private key" button.\n\n2. Save this file to somewhere on your computer. For example, presuming the previous commands downloaded a file to "$HOME/downloads/private_key.json"\n\n   .. code:: bash\n\n      mv "$HOME/downloads/private_key.json" "$HOME/.firebaseconfig/private_key.json"\n\n3. Set the environment variable GOOGLE_APPLICATION_CREDENTIALS to point to this file.\n\n   .. code:: bash\n\n      export GOOGLE_APPLICATION_CREDENTIALS="$HOME/.firebaseconfig/private_key.json"\n\n   or better yet, set this environment variable permanently in your shell configuration file.\n\n   .. code:: bash\n\n      echo "export GOOGLE_APPLICATION_CREDENTIALS=\\"$HOME/.firebaseconfig/private_key.json\\"" >> ~/.zshrc\n\n\nCommand-line Usage\n~~~~~~~~~~~~~~~~~~\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*Roarquery* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_\'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _authentication_instructions:\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/richford/roarquery/issues\n.. _Firebase project console: https://console.firebase.google.com\n.. _fuego: https://sgarciac.github.io/fuego/\n.. _fuego documentation: https://sgarciac.github.io/fuego/#installation\n.. _service account credentials: https://sgarciac.github.io/fuego/#authentication\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://roarquery.readthedocs.io/en/latest/usage.html\n',
    'author': 'Adam Richie-Halford',
    'author_email': 'richiehalford@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/richford/roarquery',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
